@@ -2,10 +2,14 @@
 #include <string.h>
 #include "Rinternals.h"
 
-SEXP set(SEXP x, PROTECT_INDEX x_idx, int i, SEXP val) {
-  if (i >= Rf_length(x)) {
-    x = Rf_lengthgets(x, i + 1);
-    REPROTECT(x, x_idx);
+SEXP set(SEXP x, int i, SEXP val) {
+  size_t len = Rf_length(x);
+  if (i >= len) {
+    // Gives us the growth sequence 3, 5, 9, 17, ...
+    // This works well because for the glue case the final number of elements
+    // will always be odd, and for common cases is 3 or 5.
+    len = (len * 2) - 1;
+    x = Rf_lengthgets(x, len);
   }
   SET_VECTOR_ELT(x, i, val);
   return x;
@@ -33,7 +37,7 @@ SEXP glue_(SEXP x, SEXP f, SEXP open_arg, SEXP close_arg) {
   const char* close = CHAR(STRING_ELT(close_arg, 0));
   size_t close_len = strlen(close);
 
-  SEXP out = Rf_allocVector(VECSXP, 1);
+  SEXP out = Rf_allocVector(VECSXP, 3);
   PROTECT_INDEX out_idx;
   PROTECT_WITH_INDEX(out, &out_idx);
 
@@ -126,24 +130,24 @@ SEXP glue_(SEXP x, SEXP f, SEXP open_arg, SEXP close_arg) {
           };
         }
         if (delim_level == 0) {
+          // Result of the current glue statement
           SEXP expr = PROTECT(Rf_ScalarString(
               Rf_mkCharLen(&xx[start], (i - close_len) + 1 - start)));
           SEXP call = PROTECT(Rf_lang2(f, expr));
           SEXP result = PROTECT(Rf_eval(call, R_GlobalEnv));
 
+          // text in between last glue statement
           str[j] = '\0';
-          out = set(out, out_idx, k++,
-                    Rf_ScalarString(Rf_mkCharLenCE(str, j, CE_UTF8)));
-          out = set(out, out_idx, k++, result);
+          SEXP str_ = PROTECT(Rf_ScalarString(Rf_mkCharLenCE(str, j, CE_UTF8)));
+          REPROTECT(out = set(out, k++, str_), out_idx);
+
+          REPROTECT(out = set(out, k++, result), out_idx);
 
           // Clear the string buffer
           memset(str, 0, j);
           j = 0;
-          UNPROTECT(3);
+          UNPROTECT(4);
           state = text;
-
-          /* Move i to end of closing delim, but it always moves by 1 */
-          /*i += close_len - 1;*/
         }
         break;
       }
@@ -151,8 +155,8 @@ SEXP glue_(SEXP x, SEXP f, SEXP open_arg, SEXP close_arg) {
   }
 
   str[j] = '\0';
-  out =
-      set(out, out_idx, k++, Rf_ScalarString(Rf_mkCharLenCE(str, j, CE_UTF8)));
+  REPROTECT(out = Rf_lengthgets(out, k + 1), out_idx);
+  SET_VECTOR_ELT(out, k, Rf_ScalarString(Rf_mkCharLenCE(str, j, CE_UTF8)));
 
   if (state == delim) {
     Rf_error("Expecting '%s'", close);
